#ifndef __SIMULATION_H__INCLUDED__
#define __SIMULATION_H__INCLUDED__

#include <graphics/SDLApplication.h>

#include <graphics/Map.h>
#include <graphics/ParticleList.h>
#include <graphics/Plant.h>
#include <graphics/Sounds.h>
#include <graphics/Sunray.h>

#include <graphics/Image.h>

#include "Environment.h"
#include "Hud.h"
#include "PlayerVehicle.h"


using namespace graphics;


class Simulation :	public Scene
{
protected:
	enum GameState
	{
		intro1,
		intro2,
		intro3,

		init_game,
		begin_game,
		playing,

		outro1,
		outro1a,
		outro2,

		end_lost,
		end_game,
		end
	};


	// required for controlling the application
	SDLApplication	*pApplication;
	GameState		gameState;
	float			fGameStateTime;
	uint_t			nQuests,
					nCurrentQuest;

	// required for controlling the appearance
	Camera			camera;

	// player + UI
	PlayerVehicle	sub;
	HUD				hud;

	// terrain
	Map				heightMap;
	Environment		env;

	// (currently fixed) models
	Model			dome1, dome2, dome3, dome4, dome5, dome6, dome7,
					harvester1, harvester2, harvester3, harvester4, harvester5;

	Sound			win,
					lose,
					intro;
	bool			bWon,
					bLost;


	// "quests"
	uint_t		nShipsDestroyed;


	// USER INPUT HANDLING
	// view
	GLfloat	distanceChange,
			phiChange,
			thetaChange;
	bool	bResetView,
			bDetonate;

	// submarine
	float		fRudderHorizontal,
				fRudderVertical,
				speed;
	int8_t		nIncreasingPower,
				nWeapon;
	float		fEnginePower,
				fIncreasingPowerTime;
	const float	fIncreasingPowerStepTime;


	// called to initialise the simulation
	virtual void	initialise();
	virtual void	load();

	// called to perform the simulation for fTime seconds
	void			_simulate_view(float time);
	void			_simulate_sub(float time);
	virtual void	simulate(float time);


	// called to load a level
	void			_load_level();

	// called to begin a quest
	void			_start_quest(const uint_t &quest);
	bool			_done_quest();


public:
	Simulation();
	virtual ~Simulation();

	inline void	application(SDLApplication *_pApplication)
	{	pApplication = _pApplication;	}

	// USER INPUT HANDLING
	// camera view distance
	inline void	zoomIn()
	{	distanceChange -= 0.2f;	}
	inline void	zoomOut()
	{	distanceChange += 0.2f;	}

	// camera angle
	inline void	rotateView(const int16_t rotateY, const int16_t rotateX)
	{	phiChange += rotateX / 300.0f;
		thetaChange += rotateY / 300.0f;	}
	inline void subDirection(const int16_t horizontal, const int16_t vertical)
	{	fRudderHorizontal += horizontal / 360.0f;
		fRudderVertical += vertical / 360.0f;	}

	// reset view
	inline void	resetView(bool bEnable)
	{	bResetView = bEnable;	}

	// submarine control
	inline void	increasingPower(const bool b)
	{	nIncreasingPower += b ? 1 : -1;
		fIncreasingPowerTime = fIncreasingPowerStepTime;	}
	inline void	decreasingPower(const bool b)
	{	nIncreasingPower += b ? -1 : 1;
		fIncreasingPowerTime = fIncreasingPowerStepTime;	}

	inline void	nextWeapon()
	{	nWeapon++;	}
	inline void	previousWeapon()
	{	nWeapon--;	}

	inline void detonate()
	{	bDetonate = true;	}
	inline void	fire(const bool bFire)
	{	sub.fire(bFire);	}


	inline void toggleTextureFading()
	{	heightMap.toggleTextureFading();	}


	// skip intro/outro pics
	inline void	skip()
	{	fGameStateTime = 0.0f;	}


	// cheats
	inline void	_win();
	inline void _lose();

	inline void	_toggleGod()
	{	sub.toggleGod();
		hud.message(sub.isGod() ? "god mode ON" : "god mode OFF");	}

	inline void	_repair()
	{	sub.repair();			}
	inline void _refill()
	{	sub.refillWeapons();	}


	void	onTargetDestroyed(const Vehicle *pVehicle);
};

#endif
