#ifndef __FRAMEWORK__INPUT_BUFFER__INCLUDED__
#define __FRAMEWORK__INPUT_BUFFER__INCLUDED__

#include <fcntl.h>
#include <io.h>

#include "Exception.h"


FRAMEWORK_BEGIN


template <class Element, unsigned int SIZE = 4096> class _InputBuffer
{
public:
	typedef _InputBuffer<Element, SIZE>	_Type;

private:
	// the input buffer must not be copied
	inline _InputBuffer<Element, SIZE>(_Type &ib)
	{	/* nothing to do */	}

	inline _Type &	operator =(_Type &ib)
	{	/* nothing to do */	}

protected:
	Element	buffer[SIZE],
			*bufferEnd,
			*bufferPos;
	int_t	hFile;
	uint_t	nPosition;
	bool	bEOF,
			bFill;

	void	_init(const char *szFileName)
	{
		hFile = open(szFileName, O_RDONLY | O_BINARY);
		if(hFile == -1)
			throw IOException(getLastErrorMessage("open() failed"));
		nPosition = 0;
		bEOF = false;
		bFill = true;
	}

	void	fill()
	{
		int_t	nRead = ::read(hFile, buffer, sizeof(Element) * SIZE);

		bufferEnd = (Element *)((char *)buffer + nRead);
		if(nRead == -1)
			throw IOException(getLastErrorMessage("read() failed"));
		if(!nRead)
			bEOF = true;

		bFill = false;
		bufferPos = buffer;
	}

	virtual void	_read(Element *pEl)
	{
		if(!bEOF && bFill)
			fill();
		if(bEOF)
			throw EndOfFileException();

		*pEl = *bufferPos;
		if(++bufferPos >= bufferEnd)
			bFill = true;
		nPosition++;
	}
	virtual int_t	_read(Element *_buffer, int_t length)
	{
		if(bEOF)
			throw EndOfFileException();

		int_t	nRead = 0,
				nCopy = 0;
		uint_t	nCopied = 0;

		if(!bFill)
		{
			nRead = (bufferEnd - bufferPos) / sizeof(Element);
			nCopy = length > nRead ? nRead : length;

			memcpy(_buffer, bufferPos, nCopy);
			length -= nCopy;
			_buffer += nCopy;
			nCopied += nCopy;
		}

		if(length <= 0)
		{
			bufferPos += nCopy;
			if(bufferPos >= bufferEnd)
				bFill = true;
		}
		else while(length > 0)
		{
			if(length >= SIZE)
			{
				nRead = ::read(hFile, _buffer, length);
				if(nRead == -1)
					throw IOException(getLastErrorMessage("read() failed"));
				if(!nRead)
				{
					bEOF = true;
					break;
				}
				else
				{
					length -= nRead;
					_buffer += nRead;
					nCopied += nRead;
					bFill = true;
				}
			}
			else
			{
				fill();
				if(bEOF)
					break;

				nRead = (bufferEnd - bufferPos) / sizeof(Element);
				nCopy = length > nRead ? nRead : length,

				memcpy(_buffer, bufferPos, nCopy);
				length -= nCopy;
				_buffer += nCopy;
				nCopied += nCopy;

				bufferPos += nCopy;
				if(bufferPos >= bufferEnd)
					bFill = true;
			}
		}

		if(!nCopied)
			throw EndOfFileException();
		nPosition += nCopied;
		return nCopied;
	}

	virtual int_t	_skip(int_t length)
	{
		if(bEOF)
			throw EndOfFileException();

		int_t	nRead = 0,
				nSkip = 0;
		uint_t	nSkipped = 0;

		if(!bFill)
		{
			nRead = (bufferEnd - bufferPos) / sizeof(Element),
			nSkip = length > nRead ? nRead : length;

			nSkipped += nSkip;
			length -= nSkip;
		}

		if(length)
		{
			int_t	nPosOld = tell(),
					nPos = lseek(hFile, length, SEEK_CUR);
			if(nPos == -1)
				throw IOException(getLastErrorMessage("seek() failed"));

			if(nPos == nPosOld)
				bEOF = true;
			else
			{
				nSkipped += nPos - nPosOld;
				bFill = true;
			}
		}
		else
		{
			bufferPos += nSkip;
			if(bufferPos >= bufferEnd)
				bFill = true;
		}

		if(!nSkipped)
			throw EndOfFileException();
		nPosition += nSkipped;
		return nSkipped;
	}

public:
	inline _InputBuffer<Element, SIZE>(const char *szFileName)
	{	_init(szFileName);	}
	inline _InputBuffer<Element, SIZE>(const string &strFileName)
	{	_init(strFileName.c_str());	}
	virtual ~_InputBuffer<Element, SIZE>()
	{	if(hFile != -1)
			close(hFile);	}

	inline Element	read()
	{
		Element	el;
		_read(&el);
		return el;
	}
	inline void		read(Element *pEl)
	{
		if(!pEl)
			throw NullPointerException(InvalidArgumentException(getErrorMessage("pEl == NULL")));
		_read(pEl);
	}
	inline void		read(Element &el)
	{	_read(&el);	}
	int_t			read(Element *_buffer, int_t length)
	{
		if(!_buffer)
			throw NullPointerException(InvalidArgumentException(getErrorMessage("_buffer == NULL")));
		if(length < 0)
			throw InvalidArgumentException(getErrorMessage("length < 0"));
		if(!length)
			return 0;

		return _read(_buffer, length);
	}

	inline int_t	skip(int_t length)
	{
		if(length < 0)
			throw InvalidArgumentException(getErrorMessage("length < 0"));
		if(!length)
			return 0;

		return _skip(length);
	}

	inline bool		eof() const
	{	return bEOF;		}
	inline int_t	tell() const
	{
		/*int_t	pos = ::tell(hFile);
		if(pos == -1)
			throw IOException(getLastErrorMessage("tell() failed"));
		return pos / sizeof(Element);*/
		return nPosition;
	}

	inline _Type &	operator >>(Element &el)
	{	read(&el);
		return *this;	}
};


class InputBuffer :	public _InputBuffer<uint8_t>
{
public:
	typedef _InputBuffer<uint8_t>	_Base;
	typedef	InputBuffer				_Type;

	inline InputBuffer(const char *szFilename) :	_Base(szFilename)
	{	/* nothing to do */	}
	inline InputBuffer(const string &strFilename) :	_Base(strFilename)
	{	/* nothing to do */	}

	inline _Type &	operator >>(int8_t &b)
	{	_read((uint8_t *)&b);
		return *this;		}
	inline _Type &	operator >>(uint8_t &b)
	{	_read(&b);
		return *this;		}
	inline _Type &	operator >>(int16_t &w)
	{	_read((uint8_t *)&w, sizeof(int16_t));
		return *this;				}
	inline _Type &	operator >>(uint16_t &w)
	{	_read((uint8_t *)&w, sizeof(uint16_t));
		return *this;				}
	inline _Type &	operator >>(int32_t &d)
	{	_read((uint8_t *)&d, sizeof(int32_t));
		return *this;				}
	inline _Type &	operator >>(uint32_t &d)
	{	_read((uint8_t *)&d, sizeof(uint32_t));
		return *this;				}
	inline _Type &	operator >>(int64_t &q)
	{	_read((uint8_t *)&q, sizeof(int64_t));
		return *this;				}
	inline _Type &	operator >>(uint64_t &q)
	{	_read((uint8_t *)&q, sizeof(uint64_t));
		return *this;				}
};


FRAMEWORK_END
#endif
