#ifndef __GRAPHICS__MATRIX_H__INCLUDED__
#define __GRAPHICS__MATRIX_H__INCLUDED__

#include <framework/Reference.h>
#include <framework/ArrayAllocator.h>

#include "graphics.h"


GRAPHICS_BEGIN


template <class NumberType, uint_t ROWS, uint_t COLUMNS> class Matrix
{
public:
	typedef Matrix<NumberType, ROWS, COLUMNS>					_Type;
	typedef _Type												_TopBase;

	typedef StaticArrayAllocator<NumberType, ROWS * COLUMNS>	Allocator;

protected:
	Allocator										alloc;
	Reference<NumberType, void, Allocator, true>	buffer;

	inline void					_copyBeforeWrite()
	{	if(buffer.getReferenceCount() > 1)
			buffer = buffer.copy();				}

	inline NumberType &			_element(uint_t index)
	{	return (buffer.getPointer())[index];	}
	inline NumberType &			element(uint_t index)
	{	_copyBeforeWrite();
		return _element(index);	}

	inline const NumberType &	getElement(uint_t index) const
	{	return (buffer.getPointer())[index];	}

public:
	inline Matrix<NumberType, ROWS, COLUMNS>() :	alloc(),
													buffer(&alloc)
	{	buffer.create();	}
	inline Matrix<NumberType, ROWS, COLUMNS>(NumberType *pMatrix) :	alloc(),
																	buffer(pMatrix, &alloc)
	{	if(buffer.isEmpty())
			buffer.create();	}
	inline Matrix<NumberType, ROWS, COLUMNS>(const NumberType *pMatrix) :	alloc(),
																			buffer(pMatrix, &alloc)
	{	if(buffer.isEmpty())
			buffer.create();	}
	inline Matrix<NumberType, ROWS, COLUMNS>(const _Type &matrix) :		buffer(matrix.buffer)
	{	/* nothing to do */	}
	virtual ~Matrix<NumberType, ROWS, COLUMNS>()
	{	/* nothing to do */	}

	inline _Type &	operator =(NumberType *pMatrix)
	{	buffer = pMatrix;
		return *this;	}
	inline _Type &	operator =(const NumberType *pMatrix)
	{	buffer = pMatrix;
		return *this;	}
	inline _Type &	operator =(const _Type &matrix)
	{
		if(&matrix != this)
			buffer = matrix.buffer;
		return *this;
	}


	inline NumberType &			_element(uint_t row, uint_t column)
	{	return _element(row * COLUMNS + column);	}
	inline NumberType &			element(uint_t row, uint_t column)
	{	_copyBeforeWrite();
		return _element(row, column);	}

	inline const NumberType &	getElement(uint_t row, uint_t column) const
	{	return getElement(row * COLUMNS + column);	}

	inline const NumberType *	getPointer() const
	{	return buffer.getPointer();	}


	void	zero()
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) = 0;
	}

	void	getTransposed(Matrix<NumberType, COLUMNS, ROWS> &transposed) const
	{
		for(uint_t row = 0; row < ROWS; row++)
			for(uint_t column = 0; column < COLUMNS; column++)
				transposed.element(column, row) = getElement(row, column);
	}
	Matrix<NumberType, COLUMNS, ROWS>	getTransposed() const
	{
		Matrix<NumberType, COLUMNS, ROWS>	transposed;
		getTransposed(transposed);
		return transposed;
	}

	template <uint_t OTHER_ROWS>
		void	getMultipliedLeft(	Matrix<NumberType, OTHER_ROWS, COLUMNS> &mResult,
									const Matrix<NumberType, OTHER_ROWS, ROWS> &mLeft	) const
	{
		mResult.zero();
		for(uint_t row = 0; row < OTHER_ROWS; row++)
			for(uint_t column = 0; column < COLUMNS; column++)
				for(uint_t i = 0; i < ROWS; i++)
					mResult._element(row, column) += mLeft.getElement(row, i) * getElement(i, column);
	}
	template <uint_t OTHER_ROWS>
		Matrix<NumberType, OTHER_ROWS, COLUMNS>
		getMultipliedLeft(const Matrix<NumberType, OTHER_ROWS, ROWS> &mLeft) const
	{	return mLeft * (*this);	}

	template <uint_t OTHER_COLUMNS>
		void	getMultipliedRight(	Matrix<NumberType, ROWS, OTHER_COLUMNS> &mResult,
									const Matrix<NumberType, COLUMNS, OTHER_COLUMNS> &mRight	) const
	{
		mResult.zero();
		for(uint_t row = 0; row < ROWS; row++)
			for(uint_t column = 0; column < OTHER_COLUMNS; column++)
				for(uint_t i = 0; i < COLUMNS; i++)
					mResult._element(row, column) += getElement(row, i) * mRight.getElement(i, column);
	}
	template <uint_t OTHER_COLUMNS>
		inline Matrix<NumberType, ROWS, OTHER_COLUMNS>
		getMultipliedRight(const Matrix<NumberType, COLUMNS, OTHER_COLUMNS> &mRight) const
	{	return (*this) * mRight;	}

	_Type &	operator +=(const _Type &matrix)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) += matrix.getElement(i);
		return *this;
	}
	_Type &	operator -=(const _Type &matrix)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) -= matrix.getElement(i);
		return *this;
	}
	_Type &	operator *=(const _Type &matrix)
	{	return *this =( *this) * matrix;	}

	_Type &	operator +=(const NumberType &scalar)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) += scalar;
		return *this;
	}
	_Type &	operator -=(const NumberType &scalar)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) -= scalar;
		return *this;
	}
	_Type &	operator *=(const NumberType &scalar)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) *= scalar;
		return *this;
	}
	_Type &	operator /=(const NumberType &scalar)
	{
		_copyBeforeWrite();
		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			_element(i) /= scalar;
		return *this;
	}

	inline _Type	operator +(_Type matrix) const
	{	return matrix += *this;	}
	inline _Type	operator -(const _Type &matrix) const
	{	_Type	_this(*this);
		return _this -= matrix;	}
	_Type			operator -() const
	{
		_Type	matrix;

		for(uint_t i = 0; i < ROWS * COLUMNS; i++)
			matrix._element(i) = -getElement(i);

		return matrix;
	}
	template <uint_t OTHER_COLUMNS>
		inline Matrix<NumberType, ROWS, OTHER_COLUMNS>
		operator *(const Matrix<NumberType, COLUMNS, OTHER_COLUMNS> &mRight) const
	{
		Matrix<NumberType, ROWS, OTHER_COLUMNS>	mResult;
		getMultipliedRight<OTHER_COLUMNS>(mResult, mRight);
		return mResult;
	}

	inline _Type	operator +(const NumberType &scalar) const
	{	_Type	matrix(*this);
		return matrix += scalar;	}
	inline _Type	operator -(const NumberType &scalar) const
	{	_Type	matrix(*this);
		return matrix -= scalar;	}
	inline _Type	operator *(const NumberType &scalar) const
	{	_Type	matrix(*this);
		return matrix *= scalar;	}
	inline _Type	operator /(const NumberType &scalar) const
	{	_Type	matrix(*this);
		return matrix /= scalar;	}
};

template <class NumberType, uint_t ROWS, uint_t COLUMNS>
	inline Matrix<NumberType, ROWS, COLUMNS>	operator +(const NumberType &scalar, const Matrix<NumberType, ROWS, COLUMNS> &matrix)
{	return matrix + scalar;	}
template <class NumberType, uint_t ROWS, uint_t COLUMNS>
	inline Matrix<NumberType, ROWS, COLUMNS>	operator -(const NumberType &scalar, const Matrix<NumberType, ROWS, COLUMNS> &matrix)
{	Matrix<NumberType, ROWS, COLUMNS>	result(-matrix);
	return result += scalar;							}
template <class NumberType, uint_t ROWS, uint_t COLUMNS>
	inline Matrix<NumberType, ROWS, COLUMNS>	operator *(const NumberType &scalar, const Matrix<NumberType, ROWS, COLUMNS> &matrix)
{	return matrix * scalar;	}


GRAPHICS_END
#endif
