#ifndef __GRAPHICS__SCENE_H__INCLUDED__
#define __GRAPHICS__SCENE_H__INCLUDED__

#include <set>
#include <vector>

#include "SDLThreadedTimer.h"
#include "SDLSemaphore.h"
#include "SDLConditional.h"

#include "Camera.h"
#include "Font.h"
#include "Texture.h"
#include "Object2.h"
#include "Object3.h"

#include "Framerate.h"


GRAPHICS_BEGIN


class Scene :	public SDLThreadedTimer
{
public:
	enum State
	{
		initialising,
		loading,
		running,
		paused,
		terminating,
		terminated
	};


protected:
	typedef	set<_Object2 *>									Object2Set;
	typedef	set<_Object3 *>									ObjectSet;
	typedef	multimap<float, Renderable *, greater<float> >	RenderMap;
	typedef	list<_Object2 *>								Object2List;
	typedef	list<_Object3 *>								ObjectList;
	typedef	set<Object *>									RemovedSet;
	typedef Vector2<float>									Vector2;
	typedef Vector3<float>									Vector;


	// simulation state
	State			state;

	// for synchronising OpenGL transformation matrices updates
	SDLSemaphore	semGlUpdate,
					semGlRender;
	SDLMutex		mObjects2,
					mObjects,
					mInserts2,
					mInserts,
					mRemoves2,
					mRemoves,
					mRemoved;

	// insert/remove objects lists
	Object2List		inserts2,
					removes2;
	ObjectList		inserts,
					removes;
	RemovedSet		removed;

	// object sets
	Object2Set		objects2;
	ObjectSet		objects,
					inserted;
	RenderMap		rm;

	// enable flags
	bool				bUpdated,
						bReload,
						bWireframe,
						bFlat,
						bFog,
						bCaustics,
						bFramerate,
						bVFCulling;

	// currently set camera
	Camera				*pCamera,
						*pGlCamera;

	// UI font used
	Font::Reference		font;

	// Scene controlled OpenGL variables
	GLfloat				fViewport[4],
						fLightAmbient[4],
						fLightDiffuse[4],
						fLightPosition[4],
						fFogColour[4],
						fFogDensity,
						fFogStart,
						fFogEnd;
	GLenum				eFogMode,
						eFogHint;

	// framerate calculator
	Framerate			framerate;


	// called to initialise the simulation
	inline void		_initialise()
	{	initialise();
		state = loading;	}
	virtual void	initialise() = 0;

	// called to load the simulation
	inline void		_load()
	{	load();
		state = running;	}
	virtual void	load();

	// called to perform the simulation for fTime seconds
	void			_simulate(float time);
	virtual void	simulate(float time) = 0;

	// called if the simulation is currently paused
	virtual void	pause(float time);

	// called to shut down the simulation
	inline void		_terminate()
	{	terminate();
		state = terminated;	}
	virtual void	terminate();


	// timer callback
	virtual void	run(uint32_t _nTimeMs);

	
	void	_glReload();

	void	glBegin2D(bool bAbsolute);
	void	glEnd2D();

	void	_setupLight();
	void	_setupFog();


public:
	Scene(Camera *_pCamera = NULL);
	virtual ~Scene();


	inline void	add(_Object2 *pObject)
	{	mInserts2.P();
		inserts2.push_back(pObject);
		mInserts2.V();				}
	inline void	add(_Object3 *pObject)
	{	mInserts.P();
		inserts.push_back(pObject);
		mInserts.V();				}
	void		add(list<_Object2 *> &l)
	{	mInserts2.P();
		inserts2.insert(inserts2.end(), l.begin(), l.end());
		mInserts2.V();										}
	void		add(list<_Object3 *> &l)
	{	mInserts.P();
		inserts.insert(inserts.end(), l.begin(), l.end());
		mInserts.V();										}
	void		add(vector<_Object2 *> &v)
	{	mInserts2.P();
		inserts2.insert(inserts2.end(), v.begin(), v.end());
		mInserts2.V();										}
	void		add(vector<_Object3 *> &v)
	{	mInserts.P();
		inserts.insert(inserts.end(), v.begin(), v.end());
		mInserts.V();										}

	inline void	remove(_Object2 *pObject)
	{	mRemoves2.P();
		removes2.push_back(pObject);
		mRemoves2.V();				}
	inline void	remove(_Object3 *pObject)
	{	mRemoves.P();
		removes.push_back(pObject);
		mRemoves.V();				}
	void		remove(list<_Object2 *> &l)
	{	mRemoves2.P();
		removes2.insert(removes2.end(), l.begin(), l.end());
		mRemoves2.V();										}
	void		remove(list<_Object3 *> &l)
	{	mRemoves.P();
		removes.insert(removes.end(), l.begin(), l.end());
		mRemoves.V();										}
	void		remove(vector<_Object2 *> &v)
	{	mRemoves2.P();
		removes2.insert(removes2.end(), v.begin(), v.end());
		mRemoves2.V();										}
	void		remove(vector<_Object3 *> &v)
	{	mRemoves.P();
		removes.insert(removes.end(), v.begin(), v.end());
		mRemoves.V();										}

	void		clear();


	inline const GLfloat &	getLeft() const
	{	return fViewport[0];	}
	inline const GLfloat &	getBottom() const
	{	return fViewport[1];	}
	inline const GLfloat &	getWidth() const
	{	return fViewport[2];	}
	inline const GLfloat &	getHeight() const
	{	return fViewport[3];	}


	void	getCollisions(list<_Object3 *> &lCollisions, _Object3 *pObject);
	void	getObjectsInRadius(list<_Object3 *> &lObjects, _Object3 *pObject, const float &radius);


	inline void enableFlatShaded(const bool enable)
	{	bFlat = enable;				}
	inline void	toggleFlatShaded()
	{	bFlat = !bFlat;				}
	inline void enableWireframe(const bool enable)
	{	bWireframe = enable;		}
	inline void	toggleWireframe()
	{	bWireframe = !bWireframe;	}
	inline void	toggleCaustics()
	{	bCaustics = !bCaustics;		}
	inline void	enableFramerate(const bool enable)
	{	bFramerate = enable;		}
	inline void	toggleFramerate()
	{	bFramerate = !bFramerate;	}
	inline void enableViewFrustumCulling(const bool enable)
	{	bVFCulling = enable;		}
	inline void	toggleViewFrustumCulling()
	{	bVFCulling = !bVFCulling;	}

	inline void	setCamera(Camera *_pCamera)
	{	pCamera = _pCamera;	}

	void		setAmbientLight(const FloatColour &c);
	void		setAmbientLight(const GLfloat &r, const GLfloat &g, const GLfloat &b, const GLfloat &a);
	void		resetAmientLight();
	void		setLight(const FloatColour &c, const Vector &direction);
	void		setLight(	const GLfloat &r, const GLfloat &g, const GLfloat &b, const GLfloat &a,
							const GLfloat &x, const GLfloat &y, const GLfloat &z					);

	void		setFog(	const GLfloat &red, const GLfloat &green, const GLfloat &blue,
						const GLfloat &density, const GLfloat &start, const GLfloat &end,
						const GLenum &mode, const GLenum &hint = GL_DONT_CARE				);
	inline void	resetFog(const GLfloat &red = 0.0f, const GLfloat &green = 0.0f, const GLfloat &blue = 0.0f)
	{	fFogColour[0] = red; fFogColour[1] = green, fFogColour[2] = blue, fFogColour[3] = 1.0f;
		bFog = false;																			}


	inline void	glReload()
	{	bReload = true;	}
	void		glRender();
};


GRAPHICS_END
#endif
