#include <framework/Timer.h>

#ifndef _WINDOWS
#include <errno.h>
#include <time.h>
#endif

#include <stdexcept>


FRAMEWORK_BEGIN


#ifdef _WINDOWS
HANDLE Timer::hTimerQueue = NULL;
#endif


Timer::Timer()
{
#ifdef _WINDOWS
	if(!hTimerQueue)
		if(!(hTimerQueue = CreateTimerQueue()))
		{
			char *	pMessageBuffer;
			string	strError = string(__FUNCTION__ ": failed to create timer queue: ");

			if(!FormatMessage(	FORMAT_MESSAGE_ALLOCATE_BUFFER | 
								FORMAT_MESSAGE_FROM_SYSTEM | 
								FORMAT_MESSAGE_IGNORE_INSERTS,
								NULL,
								GetLastError(),
								MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
								(LPTSTR)&pMessageBuffer,
								0,
								NULL										)	)
				strError += "unknown error";
			else
				strError += pMessageBuffer;

			LocalFree(pMessageBuffer);
			throw runtime_error(strError);
		}

	hTimer = NULL;
#else
	// set up timer
	sigevent	timerSignal;

	timerSignal.sigev_notify = SIGEV_SIGNAL;
	timerSignal.sigev_signo = nRealtimeSignal;
	timerSignal.sigev_value.sival_ptr = (void *)this;

	if(timer_create(CLOCK_REALTIME, &timerSignal, &timerId) == -1)
		throw runtime_error(string("Timer::Timer(): failed to create timer: ") + strerror(errno));

	// set up signal handler
	struct sigaction	sa;

	memset(&sa, 0, sizeof(sa));
	sa.sa_flags = SA_SIGINFO;
	sa.sa_sigaction = _timer_callback;
	if(sigaction(nRealtimeSignal, &sa, NULL) == -1)
		throw runtime_error(string("Timer::Timer(): failed to set timer signal handler: ") + strerror(errno));
#endif
}

Timer::~Timer()
{
#ifdef _WINDOWS
	DeleteTimerQueue(hTimerQueue);
#else
	if(timer_delete(timerId) == -1)
		throw runtime_error(string("Timer::~Timer(): failed to delete timer: ") + strerror(errno));
#endif
}

void Timer::set(const double &dDelaySeconds, const double &dIntervalSeconds)
{
#ifdef _WINDOWS
	if(hTimer)
		DeleteTimerQueueTimer(hTimerQueue, hTimer, NULL);

	if(!CreateTimerQueueTimer(&hTimer, hTimerQueue, _timer_callback, this, (DWORD)(dDelaySeconds * 1000), (DWORD)(dIntervalSeconds * 1000), WT_EXECUTEINTIMERTHREAD))
	{
		hTimer = NULL;

		char *	pMessageBuffer;
		string	strError = string(__FUNCTION__ ": failed to create timer: ");

		if(!FormatMessage(	FORMAT_MESSAGE_ALLOCATE_BUFFER | 
							FORMAT_MESSAGE_FROM_SYSTEM | 
							FORMAT_MESSAGE_IGNORE_INSERTS,
							NULL,
							GetLastError(),
							MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
							(LPTSTR)&pMessageBuffer,
							0,
							NULL										)	)
			strError += "unknown error";
		else
			strError += pMessageBuffer;

		LocalFree(pMessageBuffer);
		throw runtime_error(strError);
	}
#else
	// start timer
	itimerspec	timerSettings;

	timerSettings.it_value.tv_sec = (unsigned int)dDelaySeconds;
	timerSettings.it_value.tv_nsec = (unsigned int)((dDelaySeconds - (unsigned int)dDelaySeconds) * 1000000000.0);
	timerSettings.it_interval.tv_sec = (unsigned int)dIntervalSeconds;
	timerSettings.it_interval.tv_nsec = (unsigned int)((dIntervalSeconds - (unsigned int)dIntervalSeconds) * 1000000000.0);

	if(timer_settime(timerId, 0, &timerSettings, NULL) == -1)
		throw runtime_error(string("Timer::set(): failed to set up timer: ") + strerror(errno));
#endif
}


#ifdef _WINDOWS
VOID CALLBACK _timer_callback(PVOID lpParameter, BOOLEAN TimerOrWaitFired)
{	((Timer *)lpParameter)->onTimeout();	}
#else
void _timer_callback(int, siginfo_t *extra, void *)
{	((Timer *)extra->si_value.sival_ptr)->onTimeout();	}
#endif


FRAMEWORK_END
